@@ -32,14 +32,16 @@ let interpret_destdir ~destdir path =
 ;;
 
 let get_dirs context ~prefix_from_command_line ~from_command_line =
+  let open Fiber.O in
   let module Roots = Install.Roots in
   let prefix_from_command_line = Option.map ~f:Path.of_string prefix_from_command_line in
-  let roots =
+  let+ roots =
     match prefix_from_command_line with
-    | None -> Context.roots context
+    | None -> Memo.run (Context.roots context)
     | Some prefix ->
       Roots.opam_from_prefix prefix ~relative:Path.relative
       |> Roots.map ~f:(fun s -> Some s)
+      |> Fiber.return
   in
   let roots = Roots.first_has_priority from_command_line roots in
   let must_be_defined name v =
@@ -139,6 +141,10 @@ module Special_file = struct
   ;;
 end
 
+type copy_kind =
+  | Substitute (** Use [Artifact_substitution.copy_file]. Will scan all bytes. *)
+  | Special of Special_file.t (** Hooks to add version numbers, replace sections, etc *)
+
 type rmdir_mode =
   | Fail
   | Warn
@@ -149,7 +155,7 @@ module type File_operations = sig
     :  src:Path.t
     -> dst:Path.t
     -> executable:bool
-    -> special_file:Special_file.t option
+    -> kind:copy_kind
     -> package:Package.Name.t
     -> conf:Artifact_substitution.Conf.t
     -> unit Fiber.t
@@ -166,7 +172,7 @@ module File_ops_dry_run (Verbosity : sig
 
   let print_line fmt = print_line ~verbosity fmt
 
-  let copy_file ~src ~dst ~executable ~special_file:_ ~package:_ ~conf:_ =
+  let copy_file ~src ~dst ~executable ~kind:_ ~package:_ ~conf:_ =
     print_line
       "Copying %s to %s (executable: %b)"
       (Path.to_string_maybe_quoted src)
@@ -205,25 +211,22 @@ module File_ops_real (W : sig
     ; callback : ?version:string -> Format.formatter -> unit
     }
 
+  type copy_special_file_status =
+    | Done
+    | Use_plain_copy
+
   let copy_special_file ~src ~package ~ic ~oc ~f =
     let open Fiber.O in
-    let plain_copy () =
-      (* CR-rgrinberg: we have fast paths for copying that we aren't making use
-         of here *)
-      seek_in ic 0;
-      Io.copy_channels ic oc;
-      Fiber.return ()
-    in
     match f ic with
-    | None -> plain_copy ()
+    | None -> Fiber.return Use_plain_copy
     (* XXX should we really be catching everything here? *)
     | exception _ ->
       User_warning.emit
         ~loc:(Loc.in_file src)
         [ Pp.text "Failed to parse file, not adding version and locations information." ];
-      plain_copy ()
+      Fiber.return Use_plain_copy
     | Some { need_version; callback } ->
-      let* version =
+      let+ version =
         if need_version
         then
           let* packages =
@@ -239,7 +242,7 @@ module File_ops_real (W : sig
       let ppf = Format.formatter_of_out_channel oc in
       callback ppf ?version;
       Format.pp_print_flush ppf ();
-      Fiber.return ()
+      Done
   ;;
 
   let process_meta ic =
@@ -334,57 +337,65 @@ module File_ops_real (W : sig
     ~src
     ~dst
     ~executable
-    ~special_file
+    ~kind
     ~package
     ~(conf : Artifact_substitution.Conf.t)
     =
     let chmod = if executable then fun _ -> 0o755 else fun _ -> 0o644 in
-    match (special_file : Special_file.t option) with
-    | None -> Artifact_substitution.copy_file ~conf ~executable ~src ~dst ~chmod ()
-    | Some sf ->
-      (* CR-rgrinberg: slow copying *)
+    let plain_copy () =
+      Io.copy_file ~chmod ~src ~dst ();
+      Fiber.return ()
+    in
+    match kind with
+    | Substitute -> Artifact_substitution.copy_file ~conf ~src ~dst ~chmod ()
+    | Special sf ->
+      let open Fiber.O in
       let ic, oc = Io.setup_copy ~chmod ~src ~dst () in
-      Fiber.finalize
-        ~finally:(fun () ->
-          Io.close_both (ic, oc);
-          Fiber.return ())
-        (fun () ->
-          let f =
-            match sf with
-            | META -> process_meta
-            | Dune_package ->
-              process_dune_package
-                ~get_location:(Artifact_substitution.Conf.get_location conf)
-          in
-          copy_special_file ~src ~package ~ic ~oc ~f)
+      let* status =
+        Fiber.finalize
+          ~finally:(fun () ->
+            Io.close_both (ic, oc);
+            Fiber.return ())
+          (fun () ->
+            let f =
+              match sf with
+              | META -> process_meta
+              | Dune_package ->
+                process_dune_package
+                  ~get_location:(Artifact_substitution.Conf.get_location conf)
+            in
+            copy_special_file ~src ~package ~ic ~oc ~f)
+      in
+      (match status with
+       | Done -> Fiber.return ()
+       | Use_plain_copy -> plain_copy ())
   ;;
 
   let remove_file_if_exists dst =
     if Path.exists dst
     then (
       print_line "Deleting %s" (Path.to_string_maybe_quoted dst);
-      print_unix_error (fun () -> Path.unlink dst))
+      print_unix_error (fun () -> Path.unlink_exn dst))
   ;;
 
   let remove_dir_if_exists ~if_non_empty dir =
-    if Path.exists dir
-    then (
-      match Path.readdir_unsorted dir with
-      | Ok [] ->
-        print_line "Deleting empty directory %s" (Path.to_string_maybe_quoted dir);
-        print_unix_error (fun () -> Path.rmdir dir)
-      | Error (e, _, _) ->
-        User_message.prerr (User_error.make [ Pp.text (Unix.error_message e) ])
-      | _ ->
-        let dir = Path.to_string_maybe_quoted dir in
-        (match if_non_empty with
-         | Warn ->
-           User_message.prerr
-             (User_error.make
-                [ Pp.textf "Directory %s is not empty, cannot delete (ignoring)." dir ])
-         | Fail ->
-           User_error.raise
-             [ Pp.textf "Please delete non-empty directory %s manually." dir ]))
+    match Path.readdir_unsorted dir with
+    | Error (Unix.ENOENT, _, _) -> ()
+    | Ok [] ->
+      print_line "Deleting empty directory %s" (Path.to_string_maybe_quoted dir);
+      print_unix_error (fun () -> Path.rmdir dir)
+    | Error (e, _, _) ->
+      User_message.prerr (User_error.make [ Pp.text (Unix.error_message e) ])
+    | _ ->
+      let dir = Path.to_string_maybe_quoted dir in
+      (match if_non_empty with
+       | Warn ->
+         User_message.prerr
+           (User_error.make
+              [ Pp.textf "Directory %s is not empty, cannot delete (ignoring)." dir ])
+       | Fail ->
+         User_error.raise
+           [ Pp.textf "Please delete non-empty directory %s manually." dir ])
   ;;
 
   let mkdir_p p =
@@ -464,6 +475,56 @@ let cmd_what = function
   | Uninstall -> "uninstall"
 ;;
 
+let install_entry
+  ~ops
+  ~conf
+  ~package
+  ~dir
+  ~create_install_files
+  (entry : Path.t Install.Entry.t)
+  ~dst
+  ~verbosity
+  =
+  let module Ops = (val ops : File_operations) in
+  let open Fiber.O in
+  let special_file = Special_file.of_entry entry in
+  (match special_file with
+   | _ when not create_install_files -> Fiber.return true
+   | Some Special_file.META | Some Special_file.Dune_package -> Fiber.return true
+   | None ->
+     Artifact_substitution.test_file ~src:entry.src ()
+     >>| (function
+      | Some_substitution -> true
+      | No_substitution -> false))
+  >>= function
+  | false -> Fiber.return entry
+  | true ->
+    let+ () =
+      (match Path.is_directory dst with
+       | true -> Ops.remove_dir_if_exists ~if_non_empty:Fail dst
+       | false -> Ops.remove_file_if_exists dst);
+      print_line
+        ~verbosity
+        "%s %s"
+        (if create_install_files then "Copying to" else "Installing")
+        (Path.to_string_maybe_quoted dst);
+      Ops.mkdir_p dir;
+      let executable = Section.should_set_executable_bit entry.section in
+      let kind =
+        match special_file with
+        | Some special -> Special special
+        | None ->
+          (* CR-emillon: for most cases we could use a fast copy here, but some
+             kinds of files do need artifact substitution(at least
+             executable files and artifacts built from generated sites
+             modules), but it's too late to know without reading the file. *)
+          Substitute
+      in
+      Ops.copy_file ~src:entry.src ~dst ~executable ~kind ~package ~conf
+    in
+    Install.Entry.set_src entry dst
+;;
+
 let run
   what
   context
@@ -565,8 +626,7 @@ let run
       let entries_per_package =
         List.map install_files ~f:(fun (package, install_file) ->
           let entries =
-            Install.Entry.load_install_file install_file (fun local ->
-              Path.append_local (Path.source Path.Source.root) local)
+            Install.Entry.load_install_file install_file Path.of_local
             |> List.filter ~f:(fun (entry : Path.t Install.Entry.t) ->
               Sections.should_install sections entry.section)
           in
@@ -628,15 +688,14 @@ let run
     Fiber.sequential_iter
       install_files_by_context
       ~f:(fun (context, entries_per_package) ->
-        let roots = get_dirs context ~prefix_from_command_line ~from_command_line in
+        let* roots = get_dirs context ~prefix_from_command_line ~from_command_line in
         let conf = Artifact_substitution.Conf.of_install ~relocatable ~roots ~context in
         Fiber.sequential_iter entries_per_package ~f:(fun (package, entries) ->
           let+ entries =
-            let paths = Install.Paths.make ~relative:Path.relative ~package ~roots in
             (* CR rgrinberg: why don't we install things concurrently? *)
             Fiber.sequential_map entries ~f:(fun entry ->
-              let special_file = Special_file.of_entry entry in
               let dst =
+                let paths = Install.Paths.make ~relative:Path.relative ~package ~roots in
                 Install.Entry.relative_installed_path entry ~paths
                 |> interpret_destdir ~destdir
               in
@@ -647,40 +706,15 @@ let run
                 files_deleted_in := Path.Set.add !files_deleted_in dir;
                 Fiber.return entry
               | Install ->
-                let* copy =
-                  match special_file with
-                  | _ when not create_install_files -> Fiber.return true
-                  | Some Special_file.META | Some Special_file.Dune_package ->
-                    Fiber.return true
-                  | None ->
-                    Artifact_substitution.test_file ~src:entry.src ()
-                    >>| (function
-                     | Some_substitution -> true
-                     | No_substitution -> false)
-                in
-                (match copy with
-                 | false -> Fiber.return entry
-                 | true ->
-                   let+ () =
-                     (match Path.is_directory dst with
-                      | true -> Ops.remove_dir_if_exists ~if_non_empty:Fail dst
-                      | false -> Ops.remove_file_if_exists dst);
-                     print_line
-                       ~verbosity
-                       "%s %s"
-                       (if create_install_files then "Copying to" else "Installing")
-                       (Path.to_string_maybe_quoted dst);
-                     Ops.mkdir_p dir;
-                     let executable = Section.should_set_executable_bit entry.section in
-                     Ops.copy_file
-                       ~src:entry.src
-                       ~dst
-                       ~executable
-                       ~special_file
-                       ~package
-                       ~conf
-                   in
-                   Install.Entry.set_src entry dst))
+                install_entry
+                  ~ops:(module Ops)
+                  ~conf
+                  ~package
+                  ~dir
+                  ~create_install_files
+                  ~dst
+                  ~verbosity
+                  entry)
           in
           if create_install_files
           then (
@@ -690,7 +724,7 @@ let run
                 ~findlib_toolchain:(Context.findlib_toolchain context)
                 package
             in
-            Io.write_file (Path.source fn) (Install.Entry.gen_install_file entries))))
+            Install.Entry.gen_install_file entries |> Io.write_file (Path.source fn))))
   in
   Path.Set.to_list !files_deleted_in
   (* This [List.rev] is to ensure we process children directories before
