@@ -188,7 +188,13 @@ type failure =
 
 let label = "dune-fetch"
 
-let fetch_curl ~unpack ~checksum ~target (url : OpamUrl.t) =
+let unpack ~target ~archive =
+  Tar.extract ~archive ~target
+  >>| Result.map_error ~f:(fun () ->
+    Pp.textf "unable to extract %S" (Path.to_string archive))
+;;
+
+let with_download url checksum ~f =
   let url = OpamUrl.to_string url in
   let temp_dir = Temp.create Dir ~prefix:"dune" ~suffix:(Filename.basename url) in
   let output = Path.relative temp_dir "download" in
@@ -211,25 +217,31 @@ let fetch_curl ~unpack ~checksum ~target (url : OpamUrl.t) =
     in
     (match checksum with
      | `Mismatch m -> Fiber.return @@ Error (Checksum_mismatch m)
-     | `New _ | `Match ->
-       (match unpack with
-        | false ->
-          Io.copy_file ~src:output ~dst:target ();
-          Fiber.return @@ Ok ()
-        | true ->
-          Fiber_job.run
-            (OpamSystem.extract_job ~dir:(Path.to_string target) (Path.to_string output))
-          >>| (function
-           | None -> Ok ()
-           | Some exn ->
-             let exn =
-               User_message.make
-                 [ Pp.textf "failed to unpackage archive downloaded from %s" url
-                 ; Pp.text "reason:"
-                 ; Exn.pp exn
-                 ]
-             in
-             Error (Unavailable (Some exn)))))
+     | `New _ | `Match -> f output)
+;;
+
+let fetch_curl ~unpack:unpack_flag ~checksum ~target (url : OpamUrl.t) =
+  with_download url checksum ~f:(fun output ->
+    match unpack_flag with
+    | false ->
+      Path.mkdir_p (Path.parent_exn target);
+      Path.rename output target;
+      Fiber.return @@ Ok ()
+    | true ->
+      unpack ~target ~archive:output
+      >>| (function
+       | Ok () -> Ok ()
+       | Error msg ->
+         let exn =
+           User_message.make
+             [ Pp.textf
+                 "failed to unpackage archive downloaded from %s"
+                 (OpamUrl.to_string url)
+             ; Pp.text "reason:"
+             ; msg
+             ]
+         in
+         Error (Unavailable (Some exn))))
 ;;
 
 let fetch_others ~unpack ~checksum ~target (url : OpamUrl.t) =
@@ -264,26 +276,14 @@ let fetch_others ~unpack ~checksum ~target (url : OpamUrl.t) =
     Error (Checksum_mismatch (Checksum.of_opam_hash expected))
 ;;
 
-let fetch_git rev_store ~target (source : Opam_repo.Source.t) =
-  let commit = Opam_repo.Source.commit source in
-  let* remote =
-    let branch =
-      match commit with
-      | Some (Branch b) -> Some b
-      | _ -> None
-    in
-    Rev_store.add_repo rev_store ~source:(Opam_repo.Source.url source) ~branch
-    >>= Rev_store.Remote.update
-  in
-  (match commit with
-   | Some (Commit ref) -> Rev_store.Remote.rev_of_ref remote ~ref
-   | Some (Branch name) | Some (Tag name) -> Rev_store.Remote.rev_of_name remote ~name
-   | None ->
-     let name = Rev_store.Remote.default_branch remote in
-     Rev_store.Remote.rev_of_name remote ~name)
+let fetch_git rev_store ~target (url : OpamUrl.t) =
+  OpamUrl.resolve url rev_store
+  >>= (function
+         | Error _ as e -> Fiber.return e
+         | Ok r -> OpamUrl.fetch_revision url r rev_store)
   >>= function
-  | None -> Fiber.return @@ Error (Unavailable None)
-  | Some at_rev ->
+  | Error msg -> Fiber.return @@ Error (Unavailable (Some msg))
+  | Ok at_rev ->
     let+ res = Rev_store.At_rev.check_out at_rev ~target in
     Ok res
 ;;
@@ -312,11 +312,10 @@ let fetch ~unpack ~checksum ~target (url : OpamUrl.t) =
       Dune_stats.finish event;
       Fiber.return ())
     (fun () ->
-      (match url.backend with
-       | `http -> fetch_curl
-       | _ -> fetch_others)
-        ~unpack
-        ~checksum
-        ~target
-        url)
+      match url.backend with
+      | `git ->
+        let* rev_store = Rev_store.get in
+        fetch_git rev_store ~target url
+      | `http -> fetch_curl ~unpack ~checksum ~target url
+      | _ -> fetch_others ~unpack ~checksum ~target url)
 ;;
