@@ -5,7 +5,7 @@ let def name dyn =
   Pp.box ~indent:2 (Pp.textf "let %s = " name ++ Dyn.pp dyn)
 ;;
 
-let rule sctx ~requires_link (exes : Executables.t) =
+let rule sctx ~requires_link =
   let open Action_builder.O in
   let* () = Action_builder.return () in
   let* locals, externals =
@@ -15,6 +15,9 @@ let rule sctx ~requires_link (exes : Executables.t) =
       | Some x -> Left x
       | None -> Right lib)
   in
+  let windows_system_values = [ "win32"; "win64"; "mingw"; "mingw64" ] in
+  let win_build_flags = [ "-ccopt"; "-D_UNICODE"; "-ccopt"; "-DUNICODE" ] in
+  let build_flags = List.map windows_system_values ~f:(fun f -> f, win_build_flags) in
   let link_flags =
     let win_link_flags =
       [ "-cclib"; "-lshell32"; "-cclib"; "-lole32"; "-cclib"; "-luuid" ]
@@ -22,12 +25,9 @@ let rule sctx ~requires_link (exes : Executables.t) =
     (* additional link flags keyed by the platform *)
     [ ( "macosx"
       , [ "-cclib"; "-framework CoreFoundation"; "-cclib"; "-framework CoreServices" ] )
-    ; "win32", win_link_flags
-    ; "win64", win_link_flags
-    ; "mingw", win_link_flags
-    ; "mingw64", win_link_flags
-    ; "beos", [ "-cclib"; "-lbsd" ] (* flags for Haiku *)
     ]
+    @ List.map windows_system_values ~f:(fun w -> w, win_link_flags)
+    @ [ "beos", [ "-cclib"; "-lbsd" ] (* flags for Haiku *) ]
   in
   let+ locals =
     Memo.parallel_map locals ~f:(fun x ->
@@ -66,13 +66,6 @@ let rule sctx ~requires_link (exes : Executables.t) =
        (Pp.concat
           ~sep:Pp.cut
           [ def
-              "executables"
-              (List
-                 (* @@DRA Want to be using the public_name here, not the
-                    internal name *)
-                 (List.map ~f:(fun (_, x) -> Dyn.String x) exes.names))
-          ; Pp.nop
-          ; def
               "external_libraries"
               (List
                  (List.filter_map externals ~f:(fun x ->
@@ -83,6 +76,11 @@ let rule sctx ~requires_link (exes : Executables.t) =
           ; Pp.nop
           ; def "local_libraries" (List locals)
           ; Pp.nop
+          ; def
+              "build_flags"
+              (let open Dyn in
+               list (pair string (list string)) build_flags)
+          ; Pp.nop
           ; def
               "link_flags"
               (let open Dyn in
@@ -98,5 +96,5 @@ let gen_rules sctx (exes : Executables.t) ~dir ~requires_link =
       ~dir
       (Action_builder.write_file_dyn
          (Path.Build.relative dir fname)
-         (rule sctx ~requires_link exes)))
+         (rule sctx ~requires_link)))
 ;;
