@@ -190,6 +190,7 @@ module Driver = struct
           Resolve.Memo.bind (resolve x) ~f:(fun lib ->
             get ~loc lib
             >>| function
+            | Some t -> Resolve.return t
             | None ->
               Resolve.fail
                 (User_error.make
@@ -198,9 +199,8 @@ module Driver = struct
                        "%S is not a %s"
                        (Lib_name.to_string name)
                        (desc ~plural:false)
-                   ])
-            | Some t -> Resolve.return t))
-        >>| Resolve.List.map ~f:Fun.id
+                   ])))
+        >>| Resolve.List.all
       in
       { info; lib; replaces }
     ;;
@@ -285,14 +285,12 @@ module Driver = struct
   ;;
 end
 
-let ppx_exe (ctx : Context.t) ~key =
-  let build_dir = Context.build_dir ctx in
-  Path.Build.relative build_dir (".ppx/" ^ key ^ "/ppx.exe")
+let ppx_exe (ctx : Build_context.t) ~key =
+  Path.Build.relative ctx.build_dir (".ppx/" ^ key ^ "/ppx.exe")
 ;;
 
 let build_ppx_driver sctx ~scope ~target ~pps ~pp_names =
   let open Memo.O in
-  let ctx = Super_context.context sctx in
   let* driver_and_libs =
     let ( let& ) t f = Resolve.Memo.bind t ~f in
     let& pps = Resolve.Memo.lift pps in
@@ -309,10 +307,10 @@ let build_ppx_driver sctx ~scope ~target ~pps ~pp_names =
   let dir = Path.Build.parent_exn target in
   let main_module_name = Module_name.of_string_allow_invalid (Loc.none, "_ppx") in
   let module_ = Module.generated ~kind:Impl ~src_dir:dir [ main_module_name ] in
-  let ml_source =
-    Module.file ~ml_kind:Impl module_ |> Option.value_exn |> Path.as_in_build_dir_exn
-  in
   let* () =
+    let ml_source =
+      Module.file ~ml_kind:Impl module_ |> Option.value_exn |> Path.as_in_build_dir_exn
+    in
     Super_context.add_rule
       sctx
       ~dir
@@ -322,16 +320,12 @@ let build_ppx_driver sctx ~scope ~target ~pps ~pp_names =
             (let open Resolve.O in
              let+ driver, _ = driver_and_libs in
              sprintf "let () = %s ()\n" driver.info.main)))
-  in
-  let linkages = [ Exe.Linkage.native_or_custom ctx ] in
-  let program : Exe.Program.t =
-    { name = Filename.remove_extension (Path.Build.basename target)
-    ; main_module_name
-    ; loc = Loc.none
-    }
-  in
-  let obj_dir = Obj_dir.for_pp ~dir in
-  let* cctx =
+  and* linkages =
+    let ctx = Super_context.context sctx in
+    let+ ocaml = Context.ocaml ctx in
+    [ Exe.Linkage.native_or_custom ocaml ]
+  and+ cctx =
+    let obj_dir = Obj_dir.for_pp ~dir in
     let requires_compile = Resolve.map driver_and_libs ~f:snd in
     let requires_link = Memo.lazy_ (fun () -> Memo.return requires_compile) in
     let flags = Ocaml_flags.of_list [ "-g"; "-w"; "-24" ] in
@@ -351,13 +345,22 @@ let build_ppx_driver sctx ~scope ~target ~pps ~pp_names =
       ~bin_annot:false
       ()
   in
-  let+ (_ : Exe.dep_graphs) = Exe.build_and_link ~program ~linkages cctx ~promote:None in
+  let+ (_ : Exe.dep_graphs) =
+    let program : Exe.Program.t =
+      { name = Filename.remove_extension (Path.Build.basename target)
+      ; main_module_name
+      ; loc = Loc.none
+      }
+    in
+    Exe.build_and_link ~program ~linkages cctx ~promote:None
+  in
   ()
 ;;
 
 let get_rules sctx key =
   let ctx = Super_context.context sctx in
-  let exe = ppx_exe ctx ~key in
+  let build_context = Context.build_context ctx in
+  let exe = ppx_exe build_context ~key in
   let* pp_names, scope =
     match Digest.from_hex key with
     | None ->
@@ -369,7 +372,7 @@ let get_rules sctx key =
         let dir =
           match project_root with
           | None -> Context.build_dir ctx
-          | Some dir -> Path.Build.append_source (Context.build_dir ctx) dir
+          | Some dir -> Path.Build.append_source build_context.build_dir dir
         in
         Scope.DB.find_by_dir dir
       in
@@ -393,7 +396,7 @@ let ppx_driver_exe (ctx : Context.t) libs =
   let key = Digest.to_string (Key.Decoded.of_libs libs |> Key.encode) in
   (* Make sure to compile ppx.exe for the compiling host. See: #2252, #2286 and
      #3698 *)
-  ppx_exe ~key (Context.host ctx)
+  Context.host ctx >>| Context.build_context >>| ppx_exe ~key
 ;;
 
 let get_cookies ~loc ~expander ~lib_name libs =
@@ -447,9 +450,17 @@ let get_cookies ~loc ~expander ~lib_name libs =
   |> List.concat
 ;;
 
-let ppx_driver_and_flags_internal sctx ~dune_version ~loc ~expander ~lib_name ~flags libs =
+let ppx_driver_and_flags_internal
+  context
+  ~dune_version
+  ~loc
+  ~expander
+  ~lib_name
+  ~flags
+  libs
+  =
   let open Action_builder.O in
-  let* flags =
+  let+ flags =
     if dune_version <= (3, 2)
     then Action_builder.List.map ~f:(Expander.expand_str expander) flags
     else
@@ -457,24 +468,19 @@ let ppx_driver_and_flags_internal sctx ~dune_version ~loc ~expander ~lib_name ~f
       Action_builder.List.concat_map flags ~f:(Expander.expand ~mode:Many expander)
       |> Action_builder.map
            ~f:(List.map ~f:(Value.to_string ~dir:(Path.build @@ Expander.dir expander)))
-  in
-  let+ cookies = Action_builder.of_memo (get_cookies ~loc ~lib_name ~expander libs) in
-  let ppx_driver_exe =
-    let ctx = Context.host (Super_context.context sctx) in
-    ppx_driver_exe ctx libs
-  in
+  and+ cookies = Action_builder.of_memo (get_cookies ~loc ~lib_name ~expander libs)
+  and+ ppx_driver_exe = Action_builder.of_memo @@ ppx_driver_exe context libs in
   ppx_driver_exe, flags @ cookies
 ;;
 
-let ppx_driver_and_flags sctx ~lib_name ~expander ~scope ~loc ~flags pps =
+let ppx_driver_and_flags ctx ~lib_name ~expander ~scope ~loc ~flags pps =
   let open Action_builder.O in
   let* libs = Resolve.Memo.read (Lib.DB.resolve_pps (Scope.libs scope) pps) in
-  let* exe, flags =
+  let+ exe, flags =
     let dune_version = Scope.project scope |> Dune_project.dune_version in
-    ppx_driver_and_flags_internal sctx ~loc ~expander ~dune_version ~lib_name ~flags libs
-  in
-  let* libs = Resolve.Memo.read (Lib.closure libs ~linking:true) in
-  let+ driver =
+    ppx_driver_and_flags_internal ctx ~loc ~expander ~dune_version ~lib_name ~flags libs
+  and+ driver =
+    let* libs = Resolve.Memo.read (Lib.closure libs ~linking:true) in
     Action_builder.of_memo (Driver.select libs ~loc:(User_file (loc, pps)))
     >>= Resolve.read
   in
@@ -512,14 +518,15 @@ let sandbox_of_setting = function
 ;;
 
 let action_for_pp ~sandbox ~loc ~expander ~action ~src =
-  let chdir = Expander.context expander |> Context.build_dir in
-  let bindings = Pform.Map.singleton (Var Input_file) [ Value.Path (Path.build src) ] in
-  let expander = Expander.add_bindings expander ~bindings in
+  let expander =
+    let bindings = Pform.Map.singleton (Var Input_file) [ Value.Path (Path.build src) ] in
+    Expander.add_bindings expander ~bindings
+  in
   let open Action_builder.O in
   Action_builder.path (Path.build src)
   >>> Action_unexpanded.expand_no_targets
         action
-        ~chdir
+        ~chdir:(Expander.context expander |> Context_name.build_dir)
         ~loc
         ~expander
         ~deps:[]
@@ -585,6 +592,7 @@ let lint_module sctx ~sandbox ~dir ~expander ~lint ~lib_name ~scope =
             ~loc
             [ Pp.text "Staged ppx rewriters cannot be used as linters." ];
         let corrected_suffix = ".lint-corrected" in
+        let ctx = Super_context.context sctx in
         let driver_and_flags =
           Action_builder.memoize
             ~cutoff:
@@ -595,7 +603,7 @@ let lint_module sctx ~sandbox ~dir ~expander ~lint ~lib_name ~scope =
             "ppx driver and flags"
             (let* () = Action_builder.return () in
              let* exe, driver, flags =
-               ppx_driver_and_flags sctx ~expander ~loc ~lib_name ~flags ~scope pps
+               ppx_driver_and_flags ctx ~expander ~loc ~lib_name ~flags ~scope pps
              in
              let+ ppx_flags =
                driver_flags
@@ -615,9 +623,7 @@ let lint_module sctx ~sandbox ~dir ~expander ~lint ~lib_name ~scope =
                  (Path.as_in_build_dir_exn
                     (Option.value_exn (Module.file source ~ml_kind)))
                  (let* exe, flags, args = driver_and_flags in
-                  let dir =
-                    Super_context.context sctx |> Context.build_dir |> Path.build
-                  in
+                  let dir = ctx |> Context.build_dir |> Path.build in
                   Command.run'
                     ~dir
                     (Ok (Path.build exe))
@@ -686,7 +692,14 @@ let pp_one_module
             ~cutoff:(List.equal String.equal)
             "ppx command"
             (let* exe, driver, flags =
-               ppx_driver_and_flags sctx ~expander ~loc ~scope ~flags ~lib_name pps
+               ppx_driver_and_flags
+                 (Super_context.context sctx)
+                 ~expander
+                 ~loc
+                 ~scope
+                 ~flags
+                 ~lib_name
+                 pps
              in
              let* driver_flags =
                Expander.expand_and_eval_set
@@ -726,7 +739,14 @@ let pp_one_module
           "ppx driver and flags"
           (let* () = Action_builder.return () in
            let* exe, driver, flags =
-             ppx_driver_and_flags sctx ~expander ~loc ~lib_name ~flags ~scope pps
+             ppx_driver_and_flags
+               (Super_context.context sctx)
+               ~expander
+               ~loc
+               ~lib_name
+               ~flags
+               ~scope
+               pps
            in
            let+ ppx_flags =
              driver_flags
@@ -789,12 +809,10 @@ let make
   ~scope
   =
   let preprocessor_deps = preprocessor_deps @ instrumentation_deps in
+  let+ ocaml = Context.ocaml (Super_context.context sctx) in
   let preprocess =
     Module_name.Per_item.map preprocess ~f:(fun pp ->
-      Preprocess.remove_future_syntax
-        ~for_:Compiler
-        pp
-        (Super_context.context sctx |> Context.ocaml).version)
+      Preprocess.remove_future_syntax ~for_:Compiler pp ocaml.version)
   in
   let preprocessor_deps, sandbox = Dep_conf_eval.unnamed preprocessor_deps ~expander in
   let sandbox =
@@ -826,17 +844,17 @@ let make
   |> Pp_spec.make
 ;;
 
-let get_ppx_driver sctx ~loc ~expander ~scope ~lib_name ~flags pps =
+let get_ppx_driver ctx ~loc ~expander ~scope ~lib_name ~flags pps =
   let open Action_builder.O in
   let* libs = Resolve.Memo.read (Lib.DB.resolve_pps (Scope.libs scope) pps) in
   let dune_version = Scope.project scope |> Dune_project.dune_version in
-  ppx_driver_and_flags_internal sctx ~loc ~expander ~dune_version ~lib_name ~flags libs
+  ppx_driver_and_flags_internal ctx ~loc ~expander ~dune_version ~lib_name ~flags libs
 ;;
 
-let ppx_exe sctx ~scope pp =
+let ppx_exe ctx ~scope pp =
   let open Resolve.Memo.O in
-  let+ libs = Lib.DB.resolve_pps (Scope.libs scope) [ Loc.none, pp ] in
-  ppx_driver_exe sctx libs
+  let* libs = Lib.DB.resolve_pps (Scope.libs scope) [ Loc.none, pp ] in
+  ppx_driver_exe ctx libs |> Resolve.Memo.lift_memo
 ;;
 
 let pped_modules_map preprocess v =
